# Load packages
import argparse
import json
import numpy as np
import os
import sys
from typing import Any, Dict, List

import basf2 as b2
import b2test_utils as b2tu
from ROOT import Belle2
import generators as ge
import simulation as si
from modularAnalysis import inputMdstList
from tracking import add_tracking_reconstruction

sourcing_dir = os.environ.get('GNN_CLEANUP_DIR')
if not sourcing_dir:
    b2.B2ERROR("GNN_CLEANUP_DIR environment variable is not set. Please cd to gnn_cleanup and run setup.sh")
    sys.exit(1)
sys.path.append(sourcing_dir)

from src.modules.HitCleanupModule import HitCleanup
from src.modules.EvaluationModules import HitEvaluate, TrackEvaluate


def create_path(args: argparse.Namespace) -> b2.Path:
    path = b2.Path()

    # ################################################
    # # Generate or load the (MC) data
    # ################################################   
    with open('examples/testing_config.json') as f:
        config: Dict[str, Any] = json.load(f)

    if args.generate:
        b2.B2INFO("Generating the (MC) data...")
        path.add_module('Progress')
        b2.conditions.prepend_globaltag('mc_production_MC15ri_a')
        path.add_module('EventInfoSetter', evtNumList=args.n, expList=config['expList'], runList=config['runList'])
        if args.event_type == 'mu-mu+':
            ge.add_kkmc_generator(path, 'mu-mu+')
        elif args.event_type == 'b_events_mixed':
            ge.add_evtgen_generator(path=path, finalstate='mixed')

        bg_filelist_path: str = config.get('bg_filelist_path', '')
        try:
            bg_file_list: List[str] = [os.path.join(bg_filelist_path, file) for file in os.listdir(bg_filelist_path) if file.endswith('.root')]
        except FileNotFoundError:
            b2.B2ERROR(f"Background file list path '{bg_filelist_path}' not found.")
            sys.exit(1)
        si.add_simulation(path, components=['CDC'], bkgfiles=bg_file_list)
    else:
        root_path: str = config.get('load_path', '')
        inputMdstList(filelist=[root_path], path=path, environmentType='default')

        path.add_module('Gearbox')
        path.add_module('Geometry')
        path.add_module('Progress')

    # ################################################
    # # Tracking and Cleanup
    # ################################################

    #copy the tracking path from the standard path and manipulate TFCDCWireHitPreparer
    tracking_path = b2.Path() 
    add_tracking_reconstruction(tracking_path, components=['CDC'])

    for module in tracking_path.modules():

        if module.type() not in ['Geometry', 'Gearbox', 'RootInput', 'TFCDC_WireHitPreparer']:
            path.add_module(module)

        elif module.type() == 'TFCDC_WireHitPreparer':
            if args.filter == 'mva':
                path.add_module(module)
            elif args.filter in ['all', 'cuts_from_DB']:
                path.add_module(module, filter=args.filter, filterParameters={})
            elif args.filter == 'gnn':
                path.add_module(module, filter='all', filterParameters={})
                # HERE IS THE GNN HIT CLEANUP
                path.add_module(HitCleanup(args.model, GNNcut=args.GNNcut, ignorePriorFlags=True))

            # this is an evaluation on hit level
            path.add_module(HitEvaluate())

    # this is an evaluation on track level
    path.add_module(TrackEvaluate(minNHits=7))

    return path

def process_path(args: argparse.Namespace) -> np.ndarray:
    path = create_path(args)
    b2tu.safe_process(path, max_event=args.n)
    track_metrics: np.ndarray = np.load('tmp_track_metrics.npy', allow_pickle=True)
    os.remove('tmp_track_metrics.npy')
    if args.additional_hitfeatures:
        hit_metrics = np.load('tmp_hit_metrics.npy', allow_pickle=True)
        os.remove('tmp_hit_metrics.npy')
        return np.concatenate((track_metrics, hit_metrics))

    return track_metrics

if __name__ == "__main__":
    # Parse the arguments
    parser = argparse.ArgumentParser(description="Process some integers.")
    parser.add_argument('--n', type=int, help='Number of events to process. Default: 100', default=100)
    parser.add_argument('--generate', action='store_true', help='Generate the (MC) data. True: generate new MC data, False: load data from config path. Default: False')
    parser.add_argument('--event_type', type=str, help='Event type. Choose between b_events_mixed and mu-mu+. Default: mu-mu+', default='mu-mu+')
    parser.add_argument('--prefilter', type=str, help='Filter value for the TFCDC_WireHitPreparer for the not GNN cleaned case. Choose between "mva" for the mva filter, "all" for no filter or "cuts_from_DB" for DB cuts. Default: mva', default='mva')
    parser.add_argument('--model', type=str, help='Model name. Choose from available models in src/models for different background levels. Default: IN_trained_highbg.pth', default='IN_trained_highbg.pth')
    parser.add_argument('--GNNcut', type=float, help='GNN cut position between 0 and 1. A high value cuts away more hits. Default: 0.15', default=0.15)  
    parser.add_argument('--additional_hitfeatures', action='store_true', help='Include track-independent hit level metrics in the output. Default: False', default=False) 
    args = parser.parse_args()

    # run the path for hits by GNN filter and baseline
    args.filter = 'gnn'
    metrics_cleaned = process_path(args)

    args.filter = args.prefilter
    metrics_uncleaned = process_path(args)

    # write out the metrics

    metrics_names: List[str] = ["fitting charge efficiency",
                                "fitting efficiency",
                                "finding efficiency",
                                "fake rate",
                                "clone rate",
                                "hit efficiency",
                                "hit purity"]
    if args.additional_hitfeatures:
        metrics_names += ["hit efficiency (all hits)",
                          "hit fake rate (all hits)",
                          "hit background rejection (all hits)"]

    # Print the metrics in a table format
    b2.B2RESULT(f"{'Metric':<30} | {'Baseline (' + args.prefilter + ')':<25} | {'GNN Filter':<25}")
    b2.B2RESULT("-" * 80)
    for metric, not_cleaned, cleaned in zip(metrics_names, metrics_uncleaned, metrics_cleaned):
        b2.B2RESULT(f"{metric:<30} | {not_cleaned:<25} | {cleaned:<25}")
