import basf2
from ROOT import Belle2

import pandas as pd
import math
import numpy as np
from typing import Dict, List, Union

class GetCDCInfo(basf2.Module):
    """
    Module to extract and store CDC hit information for further analysis.

    Attributes:
        output_path (str): Path to save the extracted hit information in gzip format.
        hitmode (str): Mode of extracting hits; either 'cdchits' or 'cdcwirehits'.
        wireposmode (str): Specifies wire position calculation mode; 'middle_of_wire' or 'z0'.
    """
    def __init__(self,
                 output_path: str = 'cdchitinfo.gzip',
                 hitmode: str = 'cdchits',
                 wireposmode: str = 'middle_of_wire', 
                 addMCinfo: bool = False):
        
        """
        Initializes the GetCDCInfo module with output path, hit mode, and wire position mode.
        """
        
        super().__init__()
        self.output_path: str = output_path
        self.hitmode: str = hitmode
        self.wireposmode: str = wireposmode
        self.addMCinfo: bool = addMCinfo
        self.hitinfo: Dict[str, List[Union[int, float]]] = {}

        if hitmode not in ['cdchits', 'cdcwirehits']:
            raise ValueError(f"Invalid hitmode: {hitmode}. Expected 'cdchits' or 'cdcwirehits'.")
        if wireposmode not in ['middle_of_wire', 'z0']:
            raise ValueError(f"Invalid wireposmode: {wireposmode}. Expected 'middle_of_wire' or 'z0'.")

    def initialize(self) -> None:
        """
        Sets up module resources and initializes data structures for storing hit information.
        """
        # Access event metadata, CDC (wire) hit collections and CDC geometry parameters
        self.eventinfo = Belle2.PyStoreObj('EventMetaData')
        self.cdchits = Belle2.PyStoreArray('CDCHits')
        self.wire_hits_vector = Belle2.PyStoreObj("CDCWireHitVector")
        self.obj_cdcgeometrypar = Belle2.CDC.CDCGeometryPar.Instance()   

        # Define keys and initialize hit information dictionary
        keys = ['event_id',
                'hit_id',
                'superlayer',
                'layer',
                'clayer',
                'wire',
                'x',
                'y',
                'r',
                'phi',
                'tdc',
                'adc',
                'tot']

        if self.addMCinfo:
            keys += ['mc_id', 'signal', 'track_id'] 

        self.hitinfo = {key: [] for key in keys}

    def event(self) -> None:
        """
        Event loop: Processes a single event to extract and store CDC hit information.
        """

        self.hitinfo = {key: [] for key in self.hitinfo.keys()}

        # get event ID
        m_event_id: int = self.eventinfo.getEvent()
        
        # redefine cdchits for wire hit mode
        if self.hitmode == 'cdcwirehits':
            self.wire_hit_vector = self.wire_hits_vector.unwrap()
            self.cdchits = [self.wire_hit_vector.at(i).getHit() for i in range(self.wire_hit_vector.size())]

        # loop over all CDC hits in the event
        for m_hit_id, cdchit in enumerate(self.cdchits):

            # ======================================================================================
            # Get CDC hit information (x, y, r, phi, wire, layer, clayer, superlayer, tdc, adc, tot)
            # ======================================================================================

            # store all CDC hit wire information
            m_superlayer: int = cdchit.getISuperLayer() #superlayer
            m_layer: int = cdchit.getILayer() #layer in superlayer
            m_clayer: int = cdchit.getICLayer() #getICLayer (0..55)
            m_wire: int = cdchit.getIWire() #Wire number in the layer

            # get wire positions (x, y) for middle of wire or z=0 position
            if self.wireposmode == 'z0':
                wire_hit = self.wire_hit_vector.at(m_hit_id)
                m_x: float = wire_hit.getRefPos2D().x()
                m_y: float = wire_hit.getRefPos2D().y()
            else:
                # https://stash.desy.de/projects/B2/repos/basf2/browse/cdc/geometry/include/CDCGeometryPar.h#469
                m_wirepos = self.obj_cdcgeometrypar.c_Aligned 
                pos_forward = self.obj_cdcgeometrypar.wireForwardPosition(m_clayer, m_wire, m_wirepos) 
                pos_backward = self.obj_cdcgeometrypar.wireBackwardPosition(m_clayer, m_wire, m_wirepos)
                m_x = (pos_forward.x()+pos_backward.x())/2
                m_y = (pos_forward.y()+pos_backward.y())/2

            # calculate radial distance and azimuthal angle
            m_r: float = np.hypot(m_x, m_y)
            m_phi: float = np.arctan2(m_y, m_x)

            # store TDC, ADC and TOT information
            m_tdc: int = cdchit.getTDCCount() #Time-to-digital converter output
            m_adc: int = cdchit.getADCCount() #ADC count of the accumulated charge in the drift cell
            m_tot: float = cdchit.getTOT() #Time-over-threshold

            if self.addMCinfo:
                # ==========================================================================
                # Get related MC particle information per hit (mcid, signal)
                # ==========================================================================

                cdchit_mc_relation = cdchit.getRelationsWith('MCParticles')
                if cdchit_mc_relation.size() > 0:
                    mcparticle = cdchit_mc_relation[0]
                    m_mc_id: int = mcparticle.getIndex()
                    m_signal: int = 1
                else:
                    m_mc_id = np.nan
                    m_signal = 0

                # ==========================================================================
                # Get related CDCSimHit per hit (track_id)
                # ==========================================================================

                cdchit_related_simhit = cdchit.getRelationsWith('CDCSimHits')
                if cdchit_related_simhit.size()>0:
                    simhit = cdchit_related_simhit[0]
                    m_track_id: int = simhit.getTrackId()
                else:
                    m_track_id = np.nan

            # ==========================================================================
            # Write out CDChit information
            # ==========================================================================

            self.hitinfo['event_id'].append(m_event_id)
            self.hitinfo['hit_id'].append(m_hit_id)
            self.hitinfo['x'].append(m_x)
            self.hitinfo['y'].append(m_y)
            self.hitinfo['r'].append(m_r)
            self.hitinfo['phi'].append(m_phi)
            self.hitinfo['wire'].append(m_wire)
            self.hitinfo['layer'].append(m_layer)
            self.hitinfo['clayer'].append(m_clayer)
            self.hitinfo['superlayer'].append(m_superlayer)
            self.hitinfo['tdc'].append(m_tdc)
            self.hitinfo['adc'].append(m_adc)
            self.hitinfo['tot'].append(m_tot)

            if self.addMCinfo:
                self.hitinfo['mc_id'].append(m_mc_id)
                self.hitinfo['signal'].append(m_signal)
                self.hitinfo['track_id'].append(m_track_id)
            

    def terminate(self) -> None:
        """
        Finalizes the module by saving the extracted CDC hit information to a file.
        """
        if self.output_path is not None:
            hit_info = pd.DataFrame(self.hitinfo)
            hit_info.set_index(['event_id', 'hit_id'], inplace=True)
            hit_info.to_parquet(self.output_path, compression='gzip')
