import basf2 as b2
import ROOT

import numpy as np
import pandas as pd
import torch
from typing import Any, Dict

import json
import os

from .GetCDCInfoModule import GetCDCInfo
from ..utils.graphbuilding import GraphBuilding
from ..utils.preprocess import Preprocess
from ..utils.interaction_network import InteractionNetwork
from ..utils.cleanup import GraphCleaner

class HitCleanup(GetCDCInfo):
    """
    A class for cleaning up hits in the Belle II CDC data using a Graph Neural Network (GNN).

    This class utilizes the `GetCDCInfo` module to retrieve and process CDC wire hit information,
    build graphs for hit relationships, and apply a GNN model to filter out background hits.

    Attributes:
        model_path (str): Path to the GNN model file.
        GNNcut (float): Threshold for hit retention based on GNN output.
        ignorePriorFlags (bool): Whether to reset prior flags set by the TFCDCWireHitPreparer module on hits.
    """

    def __init__(self,
                 model_path: str,
                 settings_path: str = None,
                 GNNcut: float = 0.15,
                 ignorePriorFlags: bool = True,
                 addMCinfo: bool = False):
        self.GNNcut = GNNcut
        self.ignorePriorFlags = ignorePriorFlags
        self.model_path = os.path.join(os.getenv('GNN_CLEANUP_DIR', ''), 'src/models', model_path)
        self.settings_path = settings_path 
        self._model_settings = self._load_model_settings() 
        self.model = self.load_model()

        super().__init__(output_path=None,
                         hitmode='cdcwirehits',
                         wireposmode='z0',
                         addMCinfo=addMCinfo)

    def initialize(self):
        super().initialize()
        self.cdc_hits = ROOT.Belle2.PyStoreArray('CDCHits')
        self.cdc_hits.registerRelationTo(self.cdc_hits)
    
    def load_model(self) -> InteractionNetwork:
        """Load the pretrained GNN model."""

        if self.model_path.endswith('.pt'):
            model = torch.load(self.model_path)
        elif self.model_path.endswith('.pth'):
            node_dim = len(self.model_settings['graphbuilding']['node_keys'])
            edge_dim = len(self.model_settings['graphbuilding']['edge_keys'])
            hidden_dim = self.model_settings['training']['model_dim']['hidden_size']
            if not os.path.exists(self.model_path):
                raise FileNotFoundError(f"Model file not found at {self.model_path}")
            try:
                state_dict = torch.load(self.model_path)
            except Exception as e:
                raise RuntimeError(f"Failed to load the model from {self.model_path}: {e}")
            model = InteractionNetwork(node_dim, edge_dim, hidden_dim)
            model.load_state_dict(state_dict)
        return model
        
    def _load_model_settings(self) -> Dict[str, Any]:
        """Private method to load model settings from a file."""

        settings = {}

        if self.settings_path and os.path.exists(self.settings_path):
            print('Loading settings from:', self.settings_path)
            with open(self.settings_path) as f:
                settings = json.load(f)
        else:
            model_settings_path = self.model_path.replace('.pth', '_settings.json')
            if os.path.exists(model_settings_path):
                print('Loading settings from:', model_settings_path)
                with open(model_settings_path) as f:
                    settings = json.load(f)
            else:
                print('No settings file found.')
                
        return settings
      
    @property
    def model_settings(self) -> Dict[str, Any]:
        """dict: Retrieve or set the model settings."""
        return self._model_settings
    
    @model_settings.setter
    def model_settings(self, value: Dict[str, Any]) -> None:
        """Setter for the model settings."""
        self._model_settings = value

    def get_hits(self) -> pd.DataFrame:
        """Retrieve the hit information as a Pandas DataFrame."""
        hit_info_df = pd.DataFrame(self.hitinfo)
        hit_info_df.set_index(['event_id', 'hit_id'], inplace=True)
        return hit_info_df

    def build_graph(self, hit_info_df: pd.DataFrame) -> Any:
        """Build a graph representation of the hits."""
        graph = GraphBuilding(hits=hit_info_df,
                               cuts=self.model_settings['cuts'],
                               addMCinfo=self.addMCinfo,
                               **self.model_settings['graphbuilding']).build_graph()
        return graph
    

    def cleanup(self, graph: Any) -> np.ndarray:
        """Apply the GNN model to the preprocessed graphs and return the hits to keep."""
        graph_preprocessed = Preprocess(graph, **self.model_settings['preprocessing'])()
        graphcleaner = GraphCleaner(graph_preprocessed, self.model, self.GNNcut)
        hits2keep = graphcleaner.hitIDs2keep
        gnn_output_perhit = graphcleaner.output_perhit_sorted

        return hits2keep, gnn_output_perhit
    
    def mark_flags(self, hits2keep: np.ndarray, gnn_output_perhit: np.ndarray) -> None:
        """Update hit flags based on the GNN cleanup results"""

        hits2keep_set = set(hits2keep)

        for m_hit_id, wire_hit in enumerate(self.wire_hit_vector):
            keep_hit = m_hit_id in hits2keep_set

            if not keep_hit:
                # Mark hit as background
                wire_hit.getAutomatonCell().setTakenFlag()
                wire_hit.getAutomatonCell().setBackgroundFlag()
                wire_hit.getAutomatonCell().setBadADCOrTOTFlag()
            
            elif self.ignorePriorFlags:
                # Reset prior flags for retained hits if `ignorePriorFlags` is True.
                wire_hit.getAutomatonCell().setTakenFlag(False)
                wire_hit.getAutomatonCell().setBackgroundFlag(False)
                wire_hit.getAutomatonCell().setBadADCOrTOTFlag(False)

            # Add GNN output value to CDC Hits as Relation Weight to itself
            wire_hit.getHit().addRelationTo(wire_hit.getHit(), gnn_output_perhit[m_hit_id])

    def event(self) -> None:
        """
        Event loop retrieving hits, building a graph, applying the GNN model, and updating hit flags.
        """
        super().event()
        hits = self.get_hits()
        graph = self.build_graph(hits)
        hits2keep, gnn_output_perhit = self.cleanup(graph)
        self.mark_flags(hits2keep, gnn_output_perhit)

       



        
