import copy
import torch
import numpy as np

class GraphCleaner:
        """
        Class for cleaning a graph based on the output of a model and a user-defined threshold.

        This class processes the graph by removing edges and nodes that do not meet the specified 
        threshold (`cut`). 

        Args:
            graph (torch_geometric.data.Data): The input graph to be cleaned.
            model (torch.nn.Module): The model used to predict the edge weights.
            cut (float): The threshold used to remove edges.

        Output:
            graph_cleaned (torch_geometric.data.Data): The cleaned graph
        """
    
        def __init__(self, graph, model, cut):
            """
            Initialize the cleanup_graph class and perform the initial cleanup based on the provided threshold.
            """
            self.graph = graph
            self.model = model
            self.cut = cut

            self.graph_cleaned = self.cleanup_graph()
    
        def cleanup_graph(self):
            """
            Clean the graph by removing edges below the threshold and their associated nodes.

            The method uses the model's output to determine which edges to keep. Nodes not associated 
            with any remaining edges are also removed. The cleaned graph is returned with updated 
            attributes reflecting the remaining nodes and edges.
            """

            graph = copy.deepcopy(self.graph) # Create a deep copy to preserve the original graph
    
   
            # Use the model to predict edge scores
            with torch.no_grad():
                self.output = self.model(graph).detach().numpy().squeeze()

            # Map GNN output from edges to each hit
            hits = graph.edge_index.detach().numpy().flatten()
            output_flat = np.repeat(self.output, 2)
            arr = np.vstack((hits, output_flat))
            sorted_indices = np.lexsort((-arr[1], arr[0]))
            arr_sorted = arr[:, sorted_indices]
            unique_hits, unique_indices = np.unique(arr_sorted[0], return_index=True)
            output_perhit = np.zeros(len(graph.x), dtype=output_flat.dtype)
            output_perhit[unique_hits.astype(int)] = arr_sorted[1, unique_indices]

            hit_ids = graph.hitid.detach().numpy()
            output_perhit_sorted = np.zeros(graph.num_hits, dtype=output_perhit.dtype)
            output_perhit_sorted[hit_ids] = output_perhit
    
            # Identify edges to keep based on the threshold
            edges2keep = self.output >= self.cut

            # Ensure edge index is transposed if necessary
            if graph.edge_index.ndim == 2 and graph.edge_index.shape[0] == 2:
                 graph.edge_index = graph.edge_index.t()

            # Select edges and extract associated node indices
            edge_index_cleaned = graph.edge_index[edges2keep]
            hits2keep = np.unique(edge_index_cleaned)
    
            # Map old node IDs to new IDs for the cleaned graph
            mapping_dict = {old_id: new_id for new_id, old_id in enumerate(hits2keep)}
            if len(edge_index_cleaned) == 0:
                edge_index_cleaned_remapped = np.empty((0, 2), dtype=int)
            else:
                edge_index_cleaned_remapped = np.vectorize(mapping_dict.get)(edge_index_cleaned)
    
            # Update the graph attributes to reflect the cleaned nodes and edges
            graph.x = graph.x[hits2keep]
            graph.edge_index = torch.tensor(edge_index_cleaned_remapped)
            graph.edge_attr = graph.edge_attr[edges2keep]
            graph.hitid = graph.hitid[hits2keep]
            graph.num_nodes = len(hits2keep)
            graph.num_edges = edges2keep.sum()
            graph.batch = len(hits2keep)

            # Adjust MC related info only if it exists
            graph.y = graph.y[edges2keep] if hasattr(graph, 'y') and graph.y is not None else None
            graph.pid = graph.pid[edges2keep] if hasattr(graph, 'pid') else None
            graph.y_pernode = graph.y_pernode[hits2keep] if hasattr(graph, 'y_pernode') else None
            graph.n_hits_per_track = graph.n_hits_per_track[hits2keep] if hasattr(graph, 'n_hits_per_track') else None
            
    
            # Store cleaned graph and remaining Monte Carlo hits (since the hit ID itself was shuffled)
            self.graph_cleaned = graph
            self.hitIDs2keep = graph.hitid.detach().numpy()
            self.output_perhit = output_perhit
            self.output_perhit_sorted = output_perhit_sorted

            return self.graph_cleaned