import numpy as np
import numexpr as ne
import pandas as pd

import torch
from torch_geometric.data import Data
from typing import List, Dict, Optional, Tuple, Union

class CDC_Layout:
    """
    Class for storing CDC detector layout information.
    """
    def __init__(self):
        self.wires_per_superlayer: List[int] = [160, 160, 192, 224, 256, 288, 320, 352, 384]
        self.axial_layers: List[int] = [0, 2, 4, 6, 8]
        self.n_det_layers: int = 56

class GraphBuilding:
    """
    Class for constructing graph representations from CDC hit data for machine learning applications.
    
    The constructed graphs include nodes (hits) and edges (relations between hits) with associated features.
    Graphs are saved in a format suitable for use with graph neural networks.

    Args:
            hits (pd.DataFrame): Input CDC hits data.
            cuts (dict): Dictionary of cuts to apply on the hit features.
            node_keys (list): Features to include for each node.
            edge_keys (list): Features to include for each edge.
            overlap (int): Allowed hit multiplicity per node.
            remove_stereo_layers (bool): Whether to exclude stereo layers from the graph.
            min_Ncdchits (int): Minimum number of hits per track for signal hit definition.
            d_sl, d_nl, d_nnl (list): Allowed same, next and next-to-next layer wire distance.
            output_path (str): Path to save the constructed graph.
    """

    def __init__(self, 
                 hits: pd.DataFrame, 
                 cuts: Dict[str, Tuple[Optional[float], Optional[float]]] = {},
                 fullgraph: bool = True,
                 shuffle: bool = True,
                 node_keys: List[str] = ['x', 'y', 'adc', 'tdc', 'tot'], 
                 edge_keys: List[str] = ['r', 'phi', 'tdc', 'adc'],
                 overlap: int = 1, 
                 remove_stereo_layers: bool = False, 
                 min_Ncdchits: int = 4,
                 d_sl: List[int] = [-1, 1],
                 d_nl: List[int] = [0],
                 d_nnl: List[int] = [-1, 0, 1],
                 output_path: Optional[str] = None,
                 addMCinfo: bool = False,
                 **kwargs
                 ):

        """
        Initialize the GraphBuilding class.        
        """
        
        self.cuts: Dict[str, Tuple[Optional[float], Optional[float]]] = cuts
        self.fullgraph: bool = fullgraph
        self.shuffle: bool = shuffle
        self.node_keys: List[str] = node_keys
        self.edge_keys: List[str] = edge_keys
        self.overlap: int = overlap
        self.remove_stereo_layers: bool = remove_stereo_layers
        self.min_cdchits: int = min_Ncdchits
        self.d_sl: List[int] = d_sl
        self.d_nl: List[int] = d_nl
        self.d_nnl: List[int] = d_nnl
        self.output_path: Optional[str] = output_path
        self.addMCinfo: bool = addMCinfo
        self.cdc_layout: CDC_Layout = CDC_Layout()

        # Prepare hits by applying cuts and shuffling
        self.hits: pd.DataFrame = self.preprocess_hits(hits)
        self.num_hits = len(hits)
        

    def preprocess_hits(self, hits: pd.DataFrame) -> pd.DataFrame:
        """
        Filter and shuffle hits for graph construction.
        """  

        hits = self.apply_cuts(hits) 
        hits = self.shuffle_hits(hits)

        return hits
    
    def calc_dphi(self, phi1: np.ndarray, phi2: np.ndarray) -> np.ndarray:
        """
        Calculate the difference in azimuthal angle (phi) between two hits.
        """
        dphi = phi2 - phi1
        dphi[dphi > np.pi] -= 2 * np.pi
        dphi[dphi < -np.pi] += 2 * np.pi
        return dphi
    
    def shuffle_hits(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Shuffle hits to randomize their order within each event (hits are usually ordered by MC generation).
        """
        df = df.copy()
        df.index.set_names(['event_id', 'hit_id'], inplace=True)

        # Shuffle 'hit_id' values within the event
        hit_ids = np.arange(len(df))
        if self.shuffle:
            np.random.seed(42)
            np.random.shuffle(hit_ids)
        df['hit_id_new'] = hit_ids
        df['hit_id_new'] = df['hit_id_new'].astype(int)
        df = df.reset_index().set_index(['event_id', 'hit_id_new'])
        df_sorted = df.sort_index(level=['event_id', 'hit_id_new'])

        return df_sorted
    
    def apply_cuts(self, hits: pd.DataFrame) -> pd.DataFrame:
        """
        Apply user-defined cuts to filter hits.
        """  
        for key, cut in self.cuts.items():
            if key in hits.columns:
                if cut[0] is not None:
                    hits = hits[hits[key] >= cut[0]]
                if cut[1] is not None:
                    hits = hits[hits[key] <= cut[1]]
        return hits

    
    def add_nhits_per_track(self, hits: pd.DataFrame) -> pd.DataFrame:
        """
        Add the number of hits per track as a new column.
        """
        nhits = hits.groupby(['track_id'])['track_id'].transform('count')
        hits = hits.assign(n_hits_per_track=nhits)
        return hits

    def remove_short_tracks(self, hits: pd.DataFrame) -> pd.DataFrame:
        """
        Mark hits from tracks with fewer hits than the minimum threshold as background hits.
        """
        MCids = np.unique(hits['mc_id'].dropna()) #list of all MC IDs in the event

        for mcid in MCids:
            hit_from_mcid = (hits['mc_id'] == mcid)
            n_hits_from_mcid = hit_from_mcid.sum().sum()
            if n_hits_from_mcid < self.min_cdchits:
                hits.loc[hit_from_mcid, 'mc_id'] = np.nan
        return hits
    
    def remove_layers(self, hits: pd.DataFrame) -> pd.DataFrame:
        """
        Remove specific layers (e.g., stereo layers) based on user preferences.
        """
        if self.remove_stereo_layers:
            axial_layers = self.cdc_layout.axial_layers
            return hits.query(f'layer in {axial_layers}')
        else:
            return hits
        
    def get_wire_pairs_fullgraph(self, hits: pd.DataFrame) -> np.ndarray:
        """
        Constructs valid wire pairs for edges based on spatial and layer proximity by phi distance.

        The method calculates wire distances for all superlayers at once and applies constraints on 
        wire differences (dw) and layer differences (dl) to determine valid pairs.
        """

        superlayers = hits['superlayer'].values
        hits = hits[['phi', 'clayer']]

        # Create matrices for pairwise comparisons
        hits1_matrix = hits.values[:, np.newaxis].astype(np.float32)  # Shape (n_hits, 1, 2)
        hits2_matrix = hits.values[np.newaxis, :].astype(np.float32)  # Shape (1, n_hits, 2)

        diff = hits1_matrix - hits2_matrix # Pairwise differences
        dphi = diff[:, :, 0]  # Wire differences (shape: n_hits, n_hits)
        dl = diff[:, :, 1]  # Layer differences (shape: n_hits, n_hits)
        
        # Correct wire distances for periodic boundary conditions
        two_pi = np.float32(2 * np.pi)
        one_pi = np.float32(np.pi)
        dphi = ne.evaluate('where(dphi > one_pi, dphi - two_pi, where(dphi < -one_pi, dphi + two_pi, dphi))')

        # allowed distances in phi per superlayer
        next_wire_distances = np.array([0.040, 0.040, 0.033, 0.029, 0.025, 0.022, 0.020, 0.018, 0.017])
        half_wire_distances = np.ceil(next_wire_distances / 2 * 1000) / 1000
        zero_wire_distances = np.array([0.001, 0.001, 0.001, 0.001, 0.001, 0.001, 0.001, 0.001, 0.001])

        if len(self.d_sl) > 0:
            same_layer_distances = next_wire_distances * np.max(np.abs(self.d_sl))
        else:
            same_layer_distances = -np.ones(9)
        if len(self.d_nl) > 0:
            next_layer_distances = half_wire_distances + next_wire_distances * np.max(np.abs(self.d_nl))
        else:
            next_layer_distances = -np.ones(9)
        if len(self.d_nnl) > 0:
            neighbour_distances = next_wire_distances * np.max(np.abs(self.d_nnl))
            next2next_layer_distances = np.maximum(zero_wire_distances, neighbour_distances)
        else:
            next2next_layer_distances = -np.ones(9)


         # Apply conditions for valid wire and layer differences
        abs_dphi = np.abs(dphi)
        valid_indices = np.where(
            ((dl == 0) & (dphi > 0) & (dphi <= same_layer_distances[superlayers])) |
            ((dl == 1) & (abs_dphi <= next_layer_distances[superlayers])) |
            ((dl == 2) & (abs_dphi <= next2next_layer_distances[superlayers]))
        )

        # Extract valid indices and add to wire pairs
        wire_pairs = np.vstack((hits.index.to_numpy()[valid_indices[0]], hits.index.to_numpy()[valid_indices[1]])).T
        return wire_pairs
    
    def get_wire_pairs_persuperlayer(self, hits: pd.DataFrame) -> np.ndarray:
        """
        Constructs valid wire pairs for edges based on spatial and layer proximity by wire IDs.

        The method calculates wire distances within each superlayer and applies constraints on 
        wire differences (dw) and layer differences (dl) to determine valid pairs.
        """

        layer_groups = hits.groupby('superlayer')

        wire_pairs = []

        # Loop over all superlayers
        for superlayer in range(9):
            if superlayer in layer_groups.groups:
                hits = layer_groups.get_group(superlayer)[['wire', 'layer']]
            else:
                continue
            n_wires = self.cdc_layout.wires_per_superlayer[superlayer]

            # Create matrices for pairwise comparisons
            hits1_matrix = hits.values[:, np.newaxis]  # Shape (n_hits, 1, 2)
            hits2_matrix = hits.values[np.newaxis, :]  # Shape (1, n_hits, 2)

            diff = hits1_matrix - hits2_matrix # Pairwise differences
            dw = diff[:, :, 0]  # Wire differences (shape: n_hits, n_hits)
            dl = diff[:, :, 1]  # Layer differences (shape: n_hits, n_hits)

            # Correct wire distances for periodic boundary conditions
            dw = (dw + n_wires / 2) % n_wires - n_wires / 2

            # Adjust wire differences for even/odd layers
            layer1_even = hits['layer'].values[:, np.newaxis] % 2 == 0
            if superlayer == 0:
                dw[(dw < 0) & layer1_even] += 1
                dw[(dw > 0) & ~layer1_even] -= 1
            else:
                dw[(dw < 0) & ~layer1_even] += 1
                dw[(dw > 0) & layer1_even] -= 1

            # Apply conditions for valid wire and layer differences
            condition_dl_0 = np.any([(dl == 0) & (dw == d) for d in self.d_sl if d >= 0], axis=0)
            condition_dl_1 = np.any([(dl == 1) & (dw == d) for d in self.d_nl if d >= 0], axis=0)
            condition_dl_2 = np.any([(dl == 2) & (dw == d) for d in self.d_nnl if d >= 0], axis=0)

            valid_pairs = condition_dl_0 | condition_dl_1 | condition_dl_2

            # Extract valid indices and add to wire pairs
            indices = np.argwhere(valid_pairs)
            indices = np.vstack((hits.index.to_numpy()[indices[:,0]], hits.index.to_numpy()[indices[:,1]])).T

            if len(wire_pairs) == 0:
                wire_pairs = indices
            else:
                wire_pairs = np.vstack((wire_pairs, indices))

        return wire_pairs
    
    def get_edges(self, hitsevent: pd.DataFrame, wire_pairs: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Extracts edge information for each wire pair.

        The method computes differences in radial distance, azimuthal angle, and timing (TDC) 
        between hit pairs.
        """

        # Store all event hit indices and values
        id1, id2 = wire_pairs.T
        hitvalues = hitsevent.values

        # Compute edge distance attributes
        for i, edge_feature in enumerate(self.edge_keys):
            if edge_feature == 'phi':
                diff = self.calc_dphi(hitvalues[id1, i], hitvalues[id2, i])
            else:
                diff = hitvalues[id1, i] - hitvalues[id2, i]
                
            edge_attr = diff if i == 0 else np.vstack((edge_attr, diff))
        
        return wire_pairs, edge_attr

    def get_edge_MCinfo(self, hitsevent: pd.Series, wire_pairs: np.ndarray) -> Tuple[np.ndarray, List[List[Union[int, float]]]]:
        """
        Extracts edge MC information for each wire pair.

        The method determines the ground truth for each edge based on MC truth 
        information.
        """

        # Store all event hit indices and values
        id1, id2 = wire_pairs.T
        hitvalues = hitsevent.values

        # Determine edge associated MC IDs and ground truth label
        mcid1 = hitvalues[id1]
        mcid2 = hitvalues[id2]

        y = (mcid1 == mcid2).astype(int) # Ground truth (same MC ID)
        edge_mc_ids = np.where(mcid1 == mcid2, mcid1, np.nan)
        edge_mc_ids = [[item] for item in edge_mc_ids]

        return np.array(y), edge_mc_ids

    def build_graph(self) -> Optional[Data]:
        """
        Constructs a graph representation of hits for a single event.

        The method includes nodes with associated features, edges with attributes, and 
        ground truth labels. The constructed graph is saved if an output path is specified.
        """

        # Extract and preprocess hits for the first event in the dataset
        event_id = self.hits.index[0][0]   
        hits = self.hits.loc[event_id]   
        if self.addMCinfo:
            hits = self.add_nhits_per_track(hits)
            hits = self.remove_short_tracks(hits)   
        hits = self.remove_layers(hits)    

        # write out nodes
        x = hits[self.node_keys].to_numpy()
        
        # Generate wire pairs for edges
        if self.fullgraph:
            wire_pairs = self.get_wire_pairs_fullgraph(hits)
        else:
            wire_pairs = self.get_wire_pairs_persuperlayer(hits)

        # write out graph edges
        edge_index, edge_attr = self.get_edges(hits[self.edge_keys], wire_pairs) 

        if len(edge_index) == 0:
            print('No edges found in graph')
            return None
        
        # convert data to PyTorch tensors
        x = torch.tensor(x, dtype=torch.float32)
        edge_attr = torch.tensor(edge_attr, dtype=torch.float32)
        edge_index = torch.tensor(edge_index, dtype=torch.int64)
        hitid = torch.tensor(hits['hit_id'], dtype=torch.int64)
        event_id = event_id

        if len(edge_index.shape) == 1:
            edge_index = edge_index.unsqueeze(0)

        # create the graph object
        graph = Data(x=x.float(),
                    edge_attr=torch.transpose(edge_attr.float(), 0, 1),
                    edge_index=edge_index,
                    hitid=hitid,
                    num_hits=self.num_hits,
                    num_nodes=len(x),
                    num_edges=len(edge_index),
                    node_keys=self.node_keys,
                    edge_keys=self.edge_keys,
                    overlap=self.overlap,
                    event_id=event_id)

        if self.addMCinfo:
            y, edge_mc_ids = self.get_edge_MCinfo(hits['mc_id'], wire_pairs)
            pid = torch.tensor(edge_mc_ids, dtype=torch.float32)
            y = torch.tensor(y, dtype=torch.uint8)
            y_pernode = torch.tensor(hits['signal'], dtype=torch.uint8)
            n_hits_per_track = torch.tensor(hits['n_hits_per_track'], dtype=torch.float32)

            graph.pid = pid
            graph.y = y.float()
            graph.y_pernode = y_pernode.float()
            graph.n_hits_per_track = n_hits_per_track

        # save the graph if an output path is provided
        if self.output_path is not None:
            torch.save(graph, self.output_path)
            print(f'{len(graph)} graph saved to {self.output_path}')  
    
        return graph     
