import copy
from typing import Dict, List, Optional, Any

class Preprocess:
    """
    Class for preprocessing graph data by applying feature clipping, shifting, and scaling.

    The preprocessing steps are tailored to adjust node and edge attributes to be within 
    suitable numerical ranges for machine learning models. The original graph is deep-copied 
    to ensure the preprocessing is non-destructive.

    Args:
        graph (Data): Input graph object containing nodes and edges with features.
        node_attr_scales (dict): Scaling factors for node attributes.
        edge_attr_scales (dict): Scaling factors for edge attributes.
        clipping_values (dict): Min/max clipping values for node attributes.
        shifting_values (dict): Shifting values for node attributes to apply before scaling.
    """

    def __init__(self,
                 graph: Any,
                 node_attr_scales: Dict[str, float] = {"x": 0.08, "y": 0.08, "adc": 0.014, "tdc": 0.01, "tot": 0.066},
                 edge_attr_scales: Dict[str, float] = {"r": 2.5, "phi": 250, "tdc": 0.01, "adc": 0.0003, "tot": 0.1},
                 clipping_values: Dict[str, List[Optional[float]]] = {"tdc": [4200, 5200]},
                 shifting_values: Dict[str, float] = {"tdc": 4200},
                 reverse: bool = False,
                 **kwargs: Any) -> None:
        
        """
        Initialize the preprocess class by setting scaling, clipping, and shifting parameters.
        """
        
        self.graph = copy.deepcopy(graph) # Deep copy to ensure original graph is not modified
        self.node_attr_scales = node_attr_scales
        self.edge_attr_scales = edge_attr_scales
        self.clipping_values = clipping_values
        self.shifting_values = shifting_values
        self.reverse = reverse
        
    def __call__(self) -> Any:
        """
        Apply preprocessing steps (clipping, shifting, scaling) to the graph.
        """
        graph = self.graph

        # Ensure graph attributes are 2D for consistent processing
        if graph.x.dim() == 1:
            graph.x = graph.x.unsqueeze(0)
        if graph.edge_attr.dim() == 1:
            graph.edge_attr = graph.edge_attr.unsqueeze(0)

        # Apply feature preprocessing steps
        graph = self.feature_clipping(graph)
        graph = self.feature_shifting(graph)
        graph = self.feature_scaling(graph)

        return graph

    def feature_scaling(self, graph: Any) -> Any:
        """
        Scale node and edge features based on predefined scaling factors.

        Node and edge features are multiplied by their corresponding scale values to 
        normalize their ranges.
        """
        # Scale node features
        for i, key in enumerate(graph.node_keys):
            if key in self.node_attr_scales:
                scale_factor = self.node_attr_scales[key] if not self.reverse else 1/self.node_attr_scales[key]
                graph.x[:,i] = graph.x[:,i] * scale_factor

        # Scale edge features
        for i, key in enumerate(graph.edge_keys):
            if key in self.edge_attr_scales:
                scale_factor = self.edge_attr_scales[key] if not self.reverse else 1/self.edge_attr_scales[key]
                graph.edge_attr[:,i] = graph.edge_attr[:,i] * scale_factor
        return graph

    def feature_clipping(self, graph: Any) -> Any:
        """
        Clip node features to specified min/max ranges.
        """
        for i, key in enumerate(graph.node_keys):
            if key in self.clipping_values:
                clipping_min, clipping_max = self.clipping_values[key]
                if clipping_min is not None or clipping_max is not None:
                    # Apply clipping to node features
                    graph.x[:,i] = graph.x[:,i].clamp(min=clipping_min, max=clipping_max)
        return graph
    
    def feature_shifting(self, graph: Any) -> Any:
        """
        Shift node features by specified values.

        This step adjusts the baseline of certain features before scaling.
        """
        for i, key in enumerate(graph.node_keys):
            if key in self.shifting_values:
                # Apply shifting to node features
                shift_value = self.shifting_values[key] if not self.reverse else -self.shifting_values[key]
                graph.x[:,i] = graph.x[:,i] - shift_value
        return graph
    